/*     CalculiX - A 3-dimensional finite element program                   */
/*              Copyright (C) 1998-2005 Guido Dhondt                          */

/*     This program is free software; you can redistribute it and/or     */
/*     modify it under the terms of the GNU General Public License as    */
/*     published by the Free Software Foundation(version 2);    */
/*                    */

/*     This program is distributed in the hope that it will be useful,   */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of    */ 
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the      */
/*     GNU General Public License for more details.                      */

/*     You should have received a copy of the GNU General Public License */
/*     along with this program; if not, write to the Free Software       */
/*     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.         */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "CalculiX.h"

void dyna(double *co, int *nk, int *kon, int *ipkon, char *lakon, int *ne, 
	       int *nodeboun, int *ndirboun, double *xboun, int *nboun,
	       int *ipompc, int *nodempc, double *coefmpc, char *labmpc,
               int *nmpc, int *nodeforc,int *ndirforc,double *xforc, 
               int *nforc,int *nelemload, char *sideload,double *xload,
	       int *nload, 
	       int *nactdof,int *neq, int *nzl,int *icol, int *irow, 
	       int *nmethod, int *ikmpc, int *ilmpc, int *ikboun, 
	       int *ilboun,double *elcon, int *nelcon, double *rhcon, 
	       int *nrhcon,double *cocon, int *ncocon,
               double *alcon, int *nalcon, double *alzero, 
               int *ielmat,int *ielorien, int *norien, double *orab, 
               int *ntmat_,double *t0, 
	       double *t1,int *ithermal,double *prestr, int *iprestr, 
	       double *vold,int *iperturb, double *sti, int *nzs, 
	       double *tinc, double *tper, double *alpham, double *betam,
	       double *veold, char *amname, double *amta,
	       int *namta, int *nam, int *iamforc, int *iamload,
	       int *iamt1,int *jout,
	       int *kode, char *filab,double *eme, double *xforcold, 
	       double *xloadold,
               double *t1old, int *iamboun, double *xbounold, int *iexpl,
               double *plicon, int *nplicon, double *plkcon,int *nplkcon,
               double *xstate, int *npmat_, char *matname, int *mint_,
               int *ncmat_, int *nstate_, double *ener, char *jobnamec,
               double *ttime, char *set, int *nset, int *istartset,
               int *iendset, int *ialset, int *nprint, char *prlab,
               char *prset, int *nener, double *trab, 
               int *inotr, int *ntrans, double *fmpc, char *cbody, int *ibody,
               double *xbody, int *nbody, double *xbodyold){

  char fneig[132]="";

  int nev,i,j,k,idof,jinc, *inum=NULL,*ipobody=NULL,inewton;

  double *d=NULL, *z=NULL, *b=NULL, *zeta=NULL,*stiini=NULL,*vini=NULL,
    *cd=NULL, *cv=NULL, *xforcact=NULL, *xloadact=NULL,
    *t1act=NULL, *ampli=NULL, *aa=NULL, *bb=NULL, *bj=NULL, *v=NULL,
    *stn=NULL, *stx=NULL, *een=NULL, *adb=NULL,*xstiff=NULL,
    *aub=NULL, *temp_array1=NULL, *temp_array2=NULL, *aux=NULL,
    *f=NULL, *fn=NULL, *xbounact=NULL,*epn=NULL,*xstateini=NULL,
    *enern=NULL,*xstaten=NULL,*eei=NULL,*enerini=NULL,*qfn=NULL,
    *qfx=NULL, *xbodyact=NULL, *cgr=NULL;

  FILE *f1;

  strcpy(fneig,jobnamec);
  strcat(fneig,".eig");

  if((f1=fopen(fneig,"rb"))==NULL){
    printf("*ERROR: cannot open eigenvalue file for reading...");
    exit(0);
  }
  if(fread(&nev,sizeof(int),1,f1)!=1){
    printf("*ERROR reading the eigenvalue file...");
    exit(0);
  }
  
  d=NNEW(double,nev);

  if(fread(d,sizeof(double),nev,f1)!=nev){
    printf("*ERROR reading the eigenvalue file...");
    exit(0);
  }

  adb=NNEW(double,neq[1]);
  aub=NNEW(double,nzs[1]);

  if(fread(adb,sizeof(double),neq[1],f1)!=neq[1]){
    printf("*ERROR reading the eigenvalue file...");
    exit(0);
  }

  if(fread(aub,sizeof(double),nzs[1],f1)!=nzs[1]){
    printf("*ERROR reading the eigenvalue file...");
    exit(0);
  }

  z=NNEW(double,neq[1]*nev);

  if(fread(z,sizeof(double),neq[1]*nev,f1)!=neq[1]*nev){
    printf("*ERROR reading the eigenvalue file...");
    exit(0);
  }

  fclose(f1);

  zeta=NNEW(double,nev);

  /* calculating the damping coefficients*/

  for(i=0;i<nev;i++){
    if(fabs(d[i])>(1.e-10)){
      zeta[i]=(*alpham+*betam*d[i]*d[i])/(2.*d[i]);
    }
    else {
      printf("*ERROR in dyna: one of the frequencies is zero");
      exit(0);
    }
  }

  /* modal decomposition of the initial conditions */

  cd=NNEW(double,nev);
  cv=NNEW(double,nev);

  for(i=0;i<nev;i++){cd[i]=0;cv[i]=0;}

  temp_array1=NNEW(double,neq[1]);
  temp_array2=NNEW(double,neq[1]);
  for(i=0;i<neq[1];i++){temp_array1[i]=0;temp_array2[i]=0;}

  /* displacement initial conditions */

  for(i=0;i<*nk;i++){
    for(j=0;j<4;j++){
      if(nactdof[4*i+j]!=0){
	idof=nactdof[4*i+j]-1;
	temp_array1[idof]=vold[4*i+j];
      }
    }
  }

  FORTRAN(op,(&neq[1],aux,temp_array1,temp_array2,adb,aub,icol,irow,nzl));

  for(i=0;i<neq[1];i++){
    for(k=0;k<nev;k++){
      cd[k]+=z[k*neq[1]+i]*temp_array2[i];
    }
  }

  /* velocity initial conditions */

  for(i=0;i<neq[1];i++){temp_array1[i]=0;temp_array2[i]=0;}
  for(i=0;i<*nk;i++){
    for(j=0;j<4;j++){
      if(nactdof[4*i+j]!=0){
	idof=nactdof[4*i+j]-1;
	temp_array1[idof]=veold[4*i+j];
      }
    }
  }
  
  FORTRAN(op,(&neq[1],aux,temp_array1,temp_array2,adb,aub,icol,irow,nzl));
  
  for(i=0;i<neq[1];i++){
    for(k=0;k<nev;k++){
      cv[k]+=z[k*neq[1]+i]*temp_array2[i];
    }
  }

  free(temp_array1);free(temp_array2);

  xforcact=NNEW(double,*nforc);
  xloadact=NNEW(double,2**nload);
  xbodyact=NNEW(double,7**nbody);
  /* copying the rotation axis and/or acceleration vector */
  for(k=0;k<7**nbody;k++){xbodyact[k]=xbody[k];}
  xbounact=NNEW(double,*nboun);
  if(*ithermal==1) t1act=NNEW(double,*nk);
  
  /* assigning the body forces to the elements */ 

  ipobody=NNEW(int,3**ne);
  FORTRAN(bodyforce,(cbody,ibody,ipobody,nbody,set,istartset,
		    iendset,ialset,&inewton,nset));
	  
  jinc=(*tper/(*tinc)/(*jout))**jout;
  if(jinc**tinc<*tper) jinc++;
	       
  b=NNEW(double,neq[1]); /* load rhs vector */
  bj=NNEW(double,nev); /* response modal decomposition */
  ampli=NNEW(double,*nam); /* instantaneous amplitude */

  /* constant coefficient of the linear amplitude function */
  aa=NNEW(double,(jinc+1)*nev); 
  /* linear coefficient of the linear amplitude function */
  bb=NNEW(double,(jinc+1)*nev);
  
  v=NNEW(double,4**nk);
  fn=NNEW(double,4**nk);
  stn=NNEW(double,6**nk);
  inum=NNEW(int,*nk);
  stx=NNEW(double,6**mint_**ne);
  if(*ithermal>1) {qfn=NNEW(double,3**nk);qfx=NNEW(double,3**mint_**ne);}

  if(strcmp1(&filab[18],"E   ")==0) een=NNEW(double,6**nk);
  if(strcmp1(&filab[36],"ENER")==0) enern=NNEW(double,*nk);

  eei=NNEW(double,6**mint_**ne);
  if(*nener==1){
    stiini=NNEW(double,6**mint_**ne);
    enerini=NNEW(double,*mint_**ne);}

  FORTRAN(dynsolv,(b,z,d,zeta,&nev,&neq[1],tinc,&jinc,jout,vold,xforcold,
           nodeforc,
	   ndirforc,xforc,iamforc,nforc,xloadold,xload,iamload,nelemload,
	   sideload,nload,t1old,
	   t1,iamt1,nk,amta,namta,nam,ampli,aa,bb,bj,v,nodeboun,ndirboun,
	   xboun,nboun,ipompc,nodempc,coefmpc,labmpc,nmpc,nactdof,
	   iperturb,nmethod,co,kon,ipkon,lakon,ne,stn,
	   stx,elcon,nelcon,rhcon,nrhcon,alcon,nalcon,alzero,
	   ielmat,ielorien,norien,orab,ntmat_,t0,
	   ithermal,kode,cv,cd,inum,prestr,iprestr,ikmpc,ilmpc,ikboun,
	   ilboun,filab,eme,een,sti,f,fn,
           xforcact,xloadact,t1act,xbounold,xbounact,
           iamboun,iexpl,plicon,nplicon,plkcon,nplkcon,
	   xstateini,xstiff,xstate,npmat_,epn,matname,mint_,ncmat_,
           nstate_,stiini,vini,ener,enern,xstaten,ttime,eei,enerini,
	   cocon,ncocon,set,nset,istartset,iendset,ialset,nprint,prlab,prset,
	   qfx,qfn,trab,inotr,ntrans,fmpc,veold,cbody,ibody,xbody,nbody,
	   xbodyold,xbodyact,ipobody,cgr));

  free(eei);
  if(*nener==1){free(stiini);free(enerini);}

  if(strcmp1(&filab[18],"E   ")==0) free(een);
  if(strcmp1(&filab[36],"ENER")==0) free(enern);
  if(*ithermal>1) {free(qfn);free(qfx);}

  /* updating the loading at the end of the step; 
     important in case the amplitude at the end of the step
     is not equal to one */

  for(k=0;k<*nboun;++k){xboun[k]=xbounact[k];}
  for(k=0;k<*nforc;++k){xforc[k]=xforcact[k];}
  for(k=0;k<2**nload;++k){xload[k]=xloadact[k];}
  for(k=0;k<7**nbody;k=k+7){xbody[k]=xbodyact[k];}
  if(*ithermal==1){
    for(k=0;k<*nk;++k){t1[k]=t1act[k];}
  }
  
  free(v);free(fn);free(stn);free(inum);free(stx);free(adb);
  free(aub);free(z);free(b);free(zeta);free(bj);free(cd);free(cv);
  free(xforcact);free(xloadact);free(xbounact);free(aa);free(bb);
  free(ampli);free(xbodyact);

  if(*ithermal==1) free(t1act);

  /**ttime=*ttime+*tper;*/

  return;
}
