/*     CalculiX - A 3-dimensional finite element program                 */
/*              Copyright (C) 1998-2005 Guido Dhondt                          */

/*     This program is free software; you can redistribute it and/or     */
/*     modify it under the terms of the GNU General Public License as    */
/*     published by the Free Software Foundation(version 2);    */
/*                                                                       */

/*     This program is distributed in the hope that it will be useful,   */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of    */ 
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the      */
/*     GNU General Public License for more details.                      */

/*     You should have received a copy of the GNU General Public License */
/*     along with this program; if not, write to the Free Software       */
/*     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.         */

#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include "CalculiX.h"

void read(char *jobnamec, char **inpcp, int *nline, int *nset){

  /*   reads and stores the input deck in inpcp; determines the
       number of sets  */

  FILE *f1[10];

  char buff[132], fninp[132]="", includefn[32]="", *inpc=NULL,
       textpart[640],*set=NULL;

  int i,j,k,n,in=0,nlinemax=100000,irestartread,irestartstep,
      icntrl,nload,nforc,nboun,nk,ne,nmpc,nalset,nmat,ntmat,npmat,
      norien,nam,nprint,mint,ntrans,ncs,namtot,ncmat,memmpc,ne1d,
      ne2d,nflow,*meminset=NULL,*rmeminset=NULL,
      ithermal,nener,nstate; 

  inpc=NNEW(char,132*nlinemax);
  *nline=0;

  /* opening the input file */

  strcpy(fninp,jobnamec);
  strcat(fninp,".inp");
  if((f1[in]=fopen(fninp,"r"))==NULL){
      printf("*ERROR in read: cannot open file %s",fninp);
      exit(0);
  }

  /* starting to read the input file */

  do{
      if(fgets(buff,132,f1[in])==NULL){
	  close(f1[in]);
	  if(in!=0){
	      in--;
	      continue;
	  }
	  else{break;}
	  
	  (*nline)++;
	  if(*nline>nlinemax){
	      nlinemax=(int)(1.1*132*nlinemax);
	      RENEW(inpc,char,132*nlinemax);
	  }
	  
	  /* storing the significant characters */
	  
	  i=-1;
	  do{
	      i++;
	      if((buff[i]=='\0')||(buff[i]=='\n')) break;
	      inpc[132**nline+i]=buff[i];
	  }while(1);
	  
	  /* filling with blanks */
	  
	  for(j=i;j<132;j++){
	      inpc[132**nline+j]=' ';
	      buff[j]=' ';
	  }
	  
	  /* get rid of blanks an convert to upper case before the
             first comma */
	  
	  k=0;
	  for(j=0;j<i;j++){
	      if(buff[j]==',') break;
	      if(buff[j]==' ') continue;
	      buff[j]=toupper(buff[j]);
	      buff[k]=buff[j];
	      k++;
	  }
	  
          /* check for include statements */
	  
	  if(strcmp1(&buff[0],"*INCLUDE")==0){
	      FORTRAN(includefilename,(buff,includefn));
	      in++;
	      if(in>9){
		  printf("*ERROR in read: include statements can \n
                          not be cascaded over more than 9 levels");
	      }
	      if((f1[in]=fopen(fninp,"r"))==NULL){
		  printf("*ERROR in read: cannot open file %s",fninp);
		  exit(0);
	      }
	  }
	  
	  /* counting sets */
	  
	  else if((strcmp1(&buff[0],"*NODE")==0)&&
	      (strcmp1(&buff[0],"*NODEPRINT")!=0)&&
	      (strcmp1(&buff[0],"*NODEFILE")!=0)){nset++;}
          else if(strcmp1(&buff[0],"*ELEMENT")==0){nset++;}
          else if(strcmp1(&buff[0],"*ELSET")==0){nset++;}
	  else if(strcmp1(&buff[0],"*NSET")==0){nset++;}
	  else if(strcmp1(&buff[0],"*SURFACE")==0){nset++;}
	  else if(strcmp1(&buff[0],"*RESTART")==0){
	      irestartread=0;
	      irestartstep=0;
	      FORTRAN(splitline,(buff,textpart,&n));
	      for(i=0;i<n;i++){
		  if(strcmp1(&textpart[40*i],"READ")==0){
		      irestartread=1;
		      if(irestartstep==0) irestartstep=1;
		  }
		  if(strcmp1(&textpart[40*i],"STEP")==0){
		  /***/
		  }
		  if(irestartread==1){
		      icntrl=0;
		      FORTRAN(restartshort,(nset,&nload,&nforc,&nboun,&nk,
			&ne,&nmpc,&nalset,&nmat,&ntmat,&npmat,&norien,&nam,
			&nprint,&mint,&ntrans,&ncs,&namtot,&ncmat,&memmpc,
			&ne1d,&ne2d,&nflow,set,meminset,rmeminset,jobnamec,
			&irestartstep,&icntrl,&ithermal,&nener,&nstate));
		  }
	      }
	  }
      }
  }while(1);

  RENEW(inpc,char,132**nline);
  *inpcp=inpc;
  
  return;

}
