/*     CalculiX - A 3-dimensional finite element program                   */
/*              Copyright (C) 1998-2005 Guido Dhondt                          */

/*     This program is free software; you can redistribute it and/or     */
/*     modify it under the terms of the GNU General Public License as    */
/*     published by the Free Software Foundation(version 2);    */
/*                    */

/*     This program is distributed in the hope that it will be useful,   */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of    */ 
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the      */
/*     GNU General Public License for more details.                      */

/*     You should have received a copy of the GNU General Public License */
/*     along with this program; if not, write to the Free Software       */
/*     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.         */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "CalculiX.h"
#include "sgi.h"

void sgi_factor(double *ad, double **aup, double *adb, double *aub, 
                double *sigma,double *b, int **icolp, int **irowp, 
                int *neq, int *nzs, int token){

  char *oocpath="/yatmp/scr1";
  int i,j,k,l,*icol=NULL,*irow=NULL,*pointers=NULL,method;
  long long ndim;
  double *au=NULL,ops=0,ooclimit=2000.;

  au=*aup;
  irow=*irowp;
  icol=*icolp;

  ndim=*neq+*nzs;

  RENEW(au,double,ndim);
  RENEW(irow,int,ndim);
  pointers=NNEW(int,*neq+1);

  k=ndim;
  l=*nzs;

  pointers[*neq]=ndim;
  for(i=*neq-1;i>=0;--i){
    for(j=0;j<icol[i];++j){
      irow[--k]=irow[--l]-1;
      au[k]=au[l]-*sigma*aub[l];
    }
    pointers[i]=--k;
    irow[k]=i;
    au[k]=ad[i]-*sigma*adb[i];
  }

  method=2;

  DPSLDLT_Ordering(token,method);
  DPSLDLT_Preprocess(token,*neq,pointers,irow,&ndim,&ops);

  if(*neq>250000){
    DPSLDLT_OOCLimit(token,ooclimit);
    DPSLDLT_OOCPath(token,oocpath);
    DPSLDLT_FactorOOC(token,*neq,pointers,irow,au);
  }
  else{
    DPSLDLT_Factor(token,*neq,pointers,irow,au);
  }

  free(pointers);

  k=-1;
  l=-1;

  for(i=0;i<=*neq-1;++i){
    k++;
    for(j=0;j<icol[i];++j){
      irow[++l]=irow[++k]+1;
    }
  }

  RENEW(au,double,*nzs);
  RENEW(irow,int,*nzs);
  

  *aup=au;
  *irowp=irow;
  *icolp=icol;

  return;
}
